@@ -4,6 +4,8 @@
 #include "Imlib_private.h"
 #include <setjmp.h>
 
+#define G_MAXINT ((int) 0x7fffffff)
+
 /*      Split the ID - damages input    */
 
 static char        *
@@ -41,13 +43,17 @@ _GetExtension(char *file)
 
 /*
  *     Make sure we don't wrap on our memory allocations
+ *     we check G_MAXINT/4 because rend.c malloc's w * h * bpp
+ *     + 3 is safety margin
  */
 
 void * _imlib_malloc_image(unsigned int w, unsigned int h)
 {
-       if( w > 32767 || h > 32767)
+       if (w <= 0 || w > 32767 ||
+           h <= 0 || h > 32767 ||
+           h >= (G_MAXINT/4 - 1) / w)
                return NULL;
-       return malloc(w * h * 3);
+       return malloc(w * h * 3 + 3);
 }
 
 #ifdef HAVE_LIBJPEG
@@ -360,7 +366,9 @@ _LoadTIFF(ImlibData * id, FILE *f, char *file, int *w, int *h, int *t)
   npix = ww * hh;
   *w = (int)ww;
   *h = (int)hh;
-  if(ww > 32767 || hh > 32767)
+  if (ww <= 0 || ww > 32767 ||
+      hh <= 0 || hh > 32767 ||
+      hh >= (G_MAXINT/sizeof(uint32)) / ww)
     {
        TIFFClose(tif);
        return NULL;
@@ -463,7 +471,7 @@ _LoadGIF(ImlibData * id, FILE *f, int *w, int *h, int *t)
 	    }
 	  *w = gif->Image.Width;
 	  *h = gif->Image.Height;
-	  if (*h > 32767 || *w > 32767)
+	  if (*h <= 0 || *h > 32767 || *w <= 0 || *w > 32767)
 	    {
 	       return NULL;
 	    }
@@ -965,7 +973,12 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
   comment = 0;
   quote = 0;
   context = 0;
+  memset(lookup, 0, sizeof(lookup));
+
   line = malloc(lsz);
+  if (!line)
+    return NULL;
+
   while (!done)
     {
       pc = c;
@@ -994,25 +1007,25 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
 		{
 		  /* Header */
 		  sscanf(line, "%i %i %i %i", w, h, &ncolors, &cpp);
-                  if (ncolors > 32766)
+                  if (ncolors <= 0 || ncolors > 32766)
 		    {
 		      fprintf(stderr, "IMLIB ERROR: XPM files wth colors > 32766 not supported\n");
 		      free(line);
 		      return NULL;
 		    }
-		  if (cpp > 5)
+		  if (cpp <= 0 || cpp > 5)
 		    {
 		      fprintf(stderr, "IMLIB ERROR: XPM files with characters per pixel > 5 not supported\n");
 		      free(line);
 		      return NULL;
 		    }
-		  if (*w > 32767)
+		  if (*w <= 0 || *w > 32767)
 		    {
 		      fprintf(stderr, "IMLIB ERROR: Image width > 32767 pixels for file\n");
 		      free(line);
 		      return NULL;
 		    }
-		  if (*h > 32767)
+		  if (*h <= 0 || *h > 32767)
 		    {
 		      fprintf(stderr, "IMLIB ERROR: Image height > 32767 pixels for file\n");
 		      free(line);
@@ -1045,11 +1058,13 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
 		    {
 		      int                 slen;
 		      int                 hascolor, iscolor;
+		      int                 space;
 
 		      iscolor = 0;
 		      hascolor = 0;
 		      tok[0] = 0;
 		      col[0] = 0;
+		      space = sizeof(col) - 1;
 		      s[0] = 0;
 		      len = strlen(line);
 		      strncpy(cmap[j].str, line, cpp);
@@ -1072,10 +1087,14 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
 				{
 				  if (k >= len)
 				    {
-				      if (col[0])
+				      if (col[0] && space > 0) {
 					strcat(col, " ");
-                                      if (strlen(col) + strlen(s) < sizeof(col))
+					space--;
+				      }
+				      if (slen <= space) {
 					strcat(col, s);
+					space -= slen;
+				      }
 				    }
 				  if (col[0])
 				    {
@@ -1105,14 +1124,22 @@ _LoadXPM(ImlibData * id, FILE *file, int *w, int *h, int *t)
 					    }
 					}
 				    }
-				  strcpy(tok, s);
+				  if (slen <= sizeof(tok))
+				    strcpy(tok, s);
 				  col[0] = 0;
+				  space = sizeof(col) - 1;
 				}
 			      else
 				{
-				  if (col[0])
+				  if (col[0] && space > 0)
+				  {
 				    strcat(col, " ");
-				  strcat(col, s);
+				    space--;
+				  }
+				  if (slen <= space) {
+				    strcat(col, s);
+				    space -= slen;
+				  }
 				}
 			    }
 			}
@@ -1341,12 +1368,12 @@ _LoadPPM(ImlibData * id, FILE * f, int *w, int *h)
 	  sscanf(s, "%i %i", w, h);
 	  a = *w;
 	  b = *h;
-	  if (a > 32767)
+	  if (a <= 0 || a > 32767)
 	    {
 	      fprintf(stderr, "IMLIB ERROR: Image width > 32767 pixels for file\n");
 	      return NULL;
 	    }
-	  if (b > 32767)
+	  if (b <= 0 || b > 32767)
 	    {
 	      fprintf(stderr, "IMLIB ERROR: Image height > 32767 pixels for file\n");
 	      return NULL;
@@ -1670,7 +1697,6 @@ ImlibImage         *
 Imlib_load_image(ImlibData * id, char *file)
 {
   int                 w, h;
-  int                 needs_conv = 1;
   unsigned char      *data;
   ImlibImage         *im;
   char                s[4096];
@@ -1714,74 +1740,38 @@ Imlib_load_image(ImlibData * id, char *file)
 
   if (ispnm(p))
     {
-      needs_conv = 0;
       fmt = 0;
     }
   else if (isjpeg(p))
     {
-#ifdef HAVE_LIBJPEG
-      needs_conv = 0;
       fmt = 2;
-#else
-      needs_conv = 1;
-      fmt = 0;
-#endif
     }
   else if (istiff(p))
     {
-#ifdef HAVE_LIBTIFF
-      needs_conv = 0;
       fmt = 3;
-#else
-      needs_conv = 1;
-      fmt = 0;
-#endif
     }
   else if (iseim(p))
     {
-      needs_conv = 0;
       eim = 1;
       fmt = 9999;
     }
   else if (isxpm(p))
     {
-      needs_conv = 0;
       fmt = 5;
     }
   else if (ispng(p))
     {
-#ifdef HAVE_LIBPNG
-      needs_conv = 0;
       fmt = 1;
-#else
-      needs_conv = 1;
-      fmt = 0;
-#endif
     }
   else if (isgif(p))
     {
-#ifdef HAVE_LIBGIF
-      needs_conv = 0;
       fmt = 4;
-#else
-      needs_conv = 1;
-      fmt = 0;
-#endif
     }
   else if (isbmp(p))
     {
-      needs_conv = 0;
       fmt = 6;
     }
-#if 0
-/* NO SHELL FALLBACKS! */
-  if ((needs_conv) && (id->fallback))
-    {
-      if (p != stdin) 
-        fclose(p);
-      p = open_helper("%C/convert %s pnm:-", fil, "rb");
-    }
-#endif
+
   trans = 0;
   if ((!eim) && (!data))
     {
@@ -1818,103 +1808,15 @@ Imlib_load_image(ImlibData * id, char *file)
 	  break;
 	}
     }
-  if ((p) && (!needs_conv))
-    {
-      if (p != stdin) 
-        fclose(p);
-    }
-  else if (p)
-    close_helper(p);
-#if 0
-  /* This is *all* fallback code.  Prone to badness.  Turn it OFF */
-  if ((!data) && (id->fallback))
-    {
-      p = open_helper("%C/convert %s pnm:-", fil, "rb");
-      if (p)
-	{
-	  data = _LoadPPM(id, p, &w, &h);
-	  close_helper(p);
-	}
-    }
-  if ((!eim) && (!data) && (id->fallback))
-    {
-      if (!strcasecmp(s, "jpeg"))
-	strcpy(cmd, "%J %s");
-      else if (!strcasecmp(s, "jpg"))
-	strcpy(cmd, "%J %s");
-      else if (!strcasecmp(s, "bmp"))
-	strcpy(cmd, "%P/bmptoppm %s");
-      else if (!strcasecmp(s, "ilbm"))
-	strcpy(cmd, "%P/ilbmtoppm %s");
-      else if (!strcasecmp(s, "ilb"))
-	strcpy(cmd, "%P/ilbmtoppm %s");
-      else if (!strcasecmp(s, "iff"))
-	strcpy(cmd, "%P/ilbmtoppm %s");
-      else if (!strcasecmp(s, "img"))
-	strcpy(cmd, "%P/imgtoppm %s");
-      else if (!strcasecmp(s, "mtv"))
-	strcpy(cmd, "%P/mtvtoppm %s");
-      else if (!strcasecmp(s, "pcx"))
-	strcpy(cmd, "%P/pcxtoppm %s");
-      else if (!strcasecmp(s, "pgm"))
-	strcpy(cmd, "%P/pgmtoppm rgb:ffff/ffff/ffff %s");
-      else if (!strcasecmp(s, "pi1"))
-	strcpy(cmd, "%P/pi1toppm %s");
-      else if (!strcasecmp(s, "pict"))
-	strcpy(cmd, "%P/picttoppm %s");
-      else if (!strcasecmp(s, "pic"))
-	strcpy(cmd, "%P/picttoppm %s");
-      else if (!strcasecmp(s, "pj"))
-	strcpy(cmd, "%P/pjtoppm %s");
-      else if (!strcasecmp(s, "qrt"))
-	strcpy(cmd, "%P/qrttoppm %s");
-      else if (!strcasecmp(s, "sld"))
-	strcpy(cmd, "%P/sldtoppm %s");
-      else if (!strcasecmp(s, "spc"))
-	strcpy(cmd, "%P/spctoppm %s");
-      else if (!strcasecmp(s, "spu"))
-	strcpy(cmd, "%P/sputoppm %s");
-      else if (!strcasecmp(s, "tga"))
-	strcpy(cmd, "%P/tgatoppm %s");
-      else if (!strcasecmp(s, "xim"))
-	strcpy(cmd, "%P/ximtoppm %s");
-      else if (!strcasecmp(s, "xpm"))
-	strcpy(cmd, "%P/xpmtoppm %s");
-      else if (!strcasecmp(s, "gif"))
-	strcpy(cmd, "%P/giftopnm %s");
-      else if (!strcasecmp(s, "rast"))
-	strcpy(cmd, "%P/rasttopnm %s");
-      else if (!strcasecmp(s, "ras"))
-	strcpy(cmd, "%P/rasttopnm %s");
-      else if (!strcasecmp(s, "sgi"))
-	strcpy(cmd, "%P/sgitopnm %s");
-      else if (!strcasecmp(s, "sir"))
-	strcpy(cmd, "%P/sirtopnm %s");
-      else if (!strcasecmp(s, "tiff"))
-	strcpy(cmd, "%P/tifftopnm %s");
-      else if (!strcasecmp(s, "tif"))
-	strcpy(cmd, "%P/tifftopnm %s");
-      else if (!strcasecmp(s, "wxd"))
-	strcpy(cmd, "%P/wxdtopnm %s");
-      else if (!strcasecmp(s, "zeiss"))
-	strcpy(cmd, "%P/zeisstopnm -ppm %s");
-      else if (!strcasecmp(s, "zei"))
-	strcpy(cmd, "%P/zeisstopnm -ppm %s");
-      else if (!strcasecmp(s, "zis"))
-	strcpy(cmd, "%P/zeisstopnm -ppm %s");
-      else
-	strcpy(cmd, "%P/anytopnm %s");
-      p = open_helper(cmd, fil, "rb");
-      if (p)
-	{
-	  data = _LoadPPM(id, p, &w, &h);
-	  close_helper(p);
-	}
-    }
-#endif
+
+    if (p != stdin) 
+      fclose(p);
+
   if ((!eim) && (!data))
     {
-      fprintf(stderr, "IMLIB ERROR: Cannot load image: %s\nAll fallbacks failed.\n", fil);
+      fprintf(stderr, "IMLIB ERROR: Cannot load image: %s\n"
+      		      "All fallbacks failed.\n", fil);
+      fprintf(stderr, "See /usr/share/doc/imlib2/README.fallback.\n");
       return NULL;
     }
     
