@@ -1496,36 +1496,55 @@ Imlib_create_image_from_xpm_data(ImlibData * id, char **data)
   context = 0;
   ptr = NULL;
   end = NULL;
+  memset(lookup, 0, sizeof(lookup));
 
   while (!done)
     {
       line = data[count++];
+      if (!line)
+        break;
+      line = strdup(line);
+      if (!line)
+        break;
+      len = strlen(line);
+      for (i = 0; i < len; ++i)
+	{
+	  c = line[i];
+	  if (c < 32)
+	    line[i] = 32;
+	  else if (c > 127)
+	    line[i] = 127;
+	}
       if (context == 0)
 	{
 	  /* Header */
 	  sscanf(line, "%i %i %i %i", &w, &h, &ncolors, &cpp);
-	  if (ncolors > 32766)
+	  if (ncolors <= 0 || ncolors > 32766)
 	    {
 	      fprintf(stderr, "IMLIB ERROR: XPM data wth colors > 32766 not supported\n");
 	      free(im);
+	      free(line);
 	      return NULL;
 	    }
-	  if (cpp > 5)
+	  if (cpp <= 0 || cpp > 5)
 	    {
 	      fprintf(stderr, "IMLIB ERROR: XPM data with characters per pixel > 5 not supported\n");
 	      free(im);
+	      free(line);
 	      return NULL;
 	    }
-	  if (w > 32767)
+	  if (w <= 0 || w > 32767)
 	    {
 	      fprintf(stderr, "IMLIB ERROR: Image width > 32767 pixels for data\n");
 	      free(im);
+	      free(line);
 	      return NULL;
 	    }
-	  if (h > 32767)
+	  if (h <= 0 || h > 32767)
 	    {
 	      fprintf(stderr, "IMLIB ERROR: Image height > 32767 pixels for data\n");
 	      free(im);
+	      free(line);
 	      return NULL;
 	    }
 	  cmap = malloc(sizeof(struct _cmap) * ncolors);
@@ -1533,6 +1552,7 @@ Imlib_create_image_from_xpm_data(ImlibData * id, char **data)
 	  if (!cmap)
 	    {
 	      free(im);
+	      free(line);
 	      return NULL;
 	    }
 	  im->rgb_width = w;
@@ -1542,6 +1562,7 @@ Imlib_create_image_from_xpm_data(ImlibData * id, char **data)
 	    {
 	      free(cmap);
 	      free(im);
+	      free(line);
 	      return NULL;
 	    }
 	  im->alpha_data = NULL;
@@ -1817,6 +1838,7 @@ Imlib_create_image_from_xpm_data(ImlibData * id, char **data)
 	}
       if ((ptr) && ((ptr - im->rgb_data) >= w * h * 3))
 	done = 1;
+      free(line);
     }
   if (!transp)
     {
@@ -2109,232 +2131,6 @@ Imlib_inlined_png_to_image(ImlibData *id, unsigned char *data, int data_size)
 static int          hpid;
 void                (*oldpiper) (int);	/* actually sighandler_t but BSD uses sig_t. */
 
-FILE               *
-open_helper(const char *instring, const char *fn, const char *mode)
-{
-  char                buf[256];	/* This is safe since our input strings
-
-				 * 
-				 * * 
-				 * * * 
-				 * * * * are bounded */
-  static char        *vec[16];
-  char               *p = strdup(instring);
-  char               *pp;
-  char               *ep;
-  int                 vn = 0;
-  int                 pid;
-  FILE               *fp = NULL;
-  char               *ofil = NULL;
-  int                 ofd = -1;
-
-  int                 pfd[2];
-
-  if (p == NULL)
-    return NULL;
-
-  if (strncmp(instring, "%Q", 2) == 0)
-    {
-      /*
-       *    Generate a quanting pipeline
-       */
-      fprintf(stderr, "Not currently supported: install ImageMagick.\n");
-      return NULL;
-    }
-  /*
-   *    Ok split the instring on spaces and translate
-   *      %C %P %F and %s
-   *
-   *      FIXME: We need to handle a format string that begins
-   *      %Q to indicate an 8bit quant in the pipeline first.
-   */
-
-  pp = p;
-
-  while (vn < 15)
-    {
-      while (*pp && isspace(*pp))
-	pp++;
-      ep = pp;
-      while (*ep && !isspace(*ep))
-	ep++;
-      if (*pp == 0)
-	break;
-      /* pp->ep is now the input string block */
-      if (*ep)
-	*ep++ = 0;
-
-      if (strcmp(pp, "%s") == 0)
-	vec[vn] = strdup(fn);
-      else if (strncmp(pp, "%P/", 3) == 0)
-	{
-#ifndef __EMX__
-	  strcpy(buf, NETPBM_PATH);
-	  strcat(buf, pp + 2);
-#else
-	  strcpy(buf, pp + 3);
-#endif
-	  if ((vec[vn] = strdup(buf)) == NULL)
-	    break;
-	}
-      else if (strncmp(pp, "%J", 3) == 0)
-	{
-	  if ((vec[vn] = strdup(DJPEG_PROG)) == NULL)
-	    break;
-	}
-      else if (strncmp(pp, "%H", 3) == 0)
-	{
-	  if ((vec[vn] = strdup(CJPEG_PROG)) == NULL)
-	    break;
-	}
-      else if (strncmp(pp, "%C/", 3) == 0)
-	{
-#ifndef __EMX__
-	  strcpy(buf, CONVERT_PATH);
-	  strcat(buf, pp + 2);
-#else
-	  strcat(buf, pp + 3);
-#endif
-	  if ((vec[vn] = strdup(buf)) == NULL)
-	    break;
-	}
-      else if (strncmp(pp, ">%s", 3) == 0)
-	{
-	  ofil = pp;
-	  vn++;
-	  pp = ep;
-	  continue;
-	}
-      else
-	{
-	  if ((vec[vn] = strdup(pp)) == NULL)
-	    break;
-	}
-      vn++;
-      pp = ep;
-    }
-
-  vec[vn] = NULL;
-
-  if (pipe(pfd) == -1)
-    goto oops;
-
-#ifdef __EMX__
-  setmode(pfd[0], O_BINARY);
-  setmode(pfd[1], O_BINARY);
-#endif
-
-  if (*mode == 'r')
-    {
-      fp = fdopen(pfd[0], "r");
-      if (fp == NULL)
-	goto oops;
-    }
-  else if (*mode == 'w')
-    {
-      fp = fdopen(pfd[1], "w");
-      if (fp == NULL)
-	goto oops;
-    }
-  else
-    goto oops;
-
-  if (ofil != NULL)
-    if ((ofd = open(ofil, O_WRONLY | O_TRUNC | O_CREAT)) == -1)
-      goto oops;
-
-#ifndef __EMX__
-  switch (pid = fork())
-    {
-    case -1:
-      break;
-    case 0:
-      signal(SIGPIPE, SIG_DFL);
-      if (*mode == 'r')
-	dup2(pfd[1], 1);
-      if (*mode == 'w')
-	{
-	  dup2(pfd[0], 0);
-	  if (ofd != -1)
-	    {
-	      dup2(ofd, 1);
-	      close(1);
-	    }
-	}
-      close(pfd[0]);
-      close(pfd[1]);
-      execv(vec[0], vec);
-      perror(vec[0]);
-      /* This MUST be _exit or we will hit the SIGPIPE
-       * handler in ways we dont want. We want our parent
-       * to flush the inherited file buffers not us. */
-      _exit(1);
-    default:
-      hpid = pid;
-      
-      if (ofd != -1)
-	close(ofd);
-      if (*mode == 'r')
-	close(pfd[1]);
-      else
-	close(pfd[0]);
-    }
-#else
-  {
-    int flag, tfd0, tfd1;
-    flag = fcntl(pfd[0], F_GETFD);
-    fcntl(pfd[0], F_SETFD, flag | FD_CLOEXEC);
-    flag = fcntl(pfd[1], F_GETFD);
-    fcntl(pfd[1], F_SETFD, flag | FD_CLOEXEC);
-    tfd0 = dup(0);
-    tfd1 = dup(1);
-    if (*mode == 'r')
-      dup2(pfd[1], 1);
-    if (*mode == 'w')
-      dup2(pfd[0], 0);
-    pid = spawnv(P_NOWAIT, vec[0], vec);
-    if (pid != -1) hpid = pid;
-    dup2(tfd0, 0);
-    dup2(tfd1, 1);
-    close(tfd0);
-    close(tfd1);
-    if (ofd != -1)
-      close(ofd);
-    if (*mode == 'r')
-      close(pfd[1]);
-    else
-      close(pfd[0]);
-  }
-#endif
-
-  for (vn = 0; vn < 16; vn++)
-    if (vec[vn])
-      free(vec[vn]);
-  oldpiper = signal(SIGPIPE, SIG_IGN);
-  return fp;
-
-oops:
-  if (ofd != -1)
-    close(ofd);
-  if (fp)
-    fclose(fp);
-  for (vn = 0; vn < 16; vn++)
-    if (vec[vn])
-      free(vec[vn]);
-  return NULL;
-}
-
-int
-close_helper(FILE * fp)
-{
-  int                 info;
-
-  fclose(fp);
-  signal(SIGPIPE, oldpiper);
-  waitpid(hpid, &info, 0);
-  return WEXITSTATUS(info);
-}
-
 static char         x_error = 0;
 
 static void
